;----------------------------------------------------------------------------;
; OS ?                                                                       ;
;----------------------------------------------------------------------------;

DP equ 0

INCLUDE "HEADER.INC"
  HDR_INFO
  HDR_START Main+DP
  HDR_TITLE "OS"

INCLUDE "STDLIB.ASM"
INCLUDE "TEXTMODE.ASM"
INCLUDE "IRQS.INC"

;----------------------------------------------------------------------------;

SECTION "Vbl irq", HOME[$40]
Vbl_irq:
  jp    Switch

;----------------------------------------------------------------------------;

_TASKLST equ $80
_TASKPTR equ $C0 ;$A0

;task structure:
;+0 flags:
;     1 in use
;     1 running
;     3 not used
;     3 internal RAM bank
;+1 ROM BANK
;+2 SP MSB
;+3 SP LSB

addtask: macro ;(pid, PC, SP, ROM_BK, SVBK)
  di
  ld    a, \5
  ldh   [_SVBK], a      ;select internal RAM bank
  ld    a, \2&$FF
  ld    [\3-2], a       ;store PC (LSB) on stack
  ld    a, \2>>8
  ld    [\3-1], a       ;store PC (MSB) on stack
  ld    a, (\3-10)>>8
  ldh   [_TASKLST+4*\1+2], a    ;stackpointer (MSB)
  ld    a, (\3-10)&$FF
  ldh   [_TASKLST+4*\1+3], a    ;stackpointer (LSB)
  ld    a, \4
  ldh   [_TASKLST+4*\1+1], a    ;ROM bank
  ld    a, $80+$40+\5
  ldh   [_TASKLST+4*\1+0], a    ;in use + running + work RAM bank
  ei
endm

;----------------------------------------------------------------------------;
; Main                                                                       ;
;----------------------------------------------------------------------------;

SECTION "Main", HOME
Main::
  call  InitText

  ld    a, _TASKLST-2
  ldh   [_TASKPTR], a           ;set task pointer
  ld    a, 0
  ldh   [_TASKLST+$00], a       ;clear tasklist
  ldh   [_TASKLST+$04], a
  ldh   [_TASKLST+$08], a
  ldh   [_TASKLST+$0C], a
  ldh   [_TASKLST+$10], a
  ldh   [_TASKLST+$14], a
  ldh   [_TASKLST+$18], a
  ldh   [_TASKLST+$1C], a

;  addtask 0,Shell,$FFFE,0,0     ;(pid, PC, SP, ROM_BK, SVBK)
;  addtask 0,$4000,$E000,2,2     ;(pid, PC, SP, ROM_BK, SVBK)

  addtask 0,$4000,$D100,1,0     ;(pid, PC, SP, ROM_BK, SVBK)
  addtask 1,$4000,$D200,1,0     ;(pid, PC, SP, ROM_BK, SVBK)
  addtask 2,$4000,$D300,1,0     ;(pid, PC, SP, ROM_BK, SVBK)
  addtask 3,$4000,$D400,1,0     ;(pid, PC, SP, ROM_BK, SVBK)
  addtask 4,$4000,$D500,1,0     ;(pid, PC, SP, ROM_BK, SVBK)
  addtask 5,$4000,$D600,1,0     ;(pid, PC, SP, ROM_BK, SVBK)
  addtask 6,$4000,$D700,1,0     ;(pid, PC, SP, ROM_BK, SVBK)
  addtask 7,$4000,$D800,1,0     ;(pid, PC, SP, ROM_BK, SVBK)
  addtask 8,$4000,$D900,1,0     ;(pid, PC, SP, ROM_BK, SVBK)
  addtask 9,$4000,$DA00,1,0     ;(pid, PC, SP, ROM_BK, SVBK)
  addtask 10,$4000,$DB00,1,0     ;(pid, PC, SP, ROM_BK, SVBK)
  addtask 11,$4000,$DC00,1,0     ;(pid, PC, SP, ROM_BK, SVBK)
  addtask 12,$4000,$DD00,1,0     ;(pid, PC, SP, ROM_BK, SVBK)
  addtask 13,$4000,$DE00,1,0     ;(pid, PC, SP, ROM_BK, SVBK)
  addtask 14,$4000,$DF00,1,0     ;(pid, PC, SP, ROM_BK, SVBK)
  addtask 15,$4000,$E000,1,0     ;(pid, PC, SP, ROM_BK, SVBK)


  ld    a, %00001
  ldh   [_IE], a
  ei
.thatsall:
  jr    .thatsall       ;don't do ne1thing, multitasking takes it over

;----------------------------------------------------------------------------;

Shell:
  inc   a
  ld    [$9800], a
  jr    Shell

;----------------------------------------------------------------------------;

Switch::
  pusha
  ld    hl, [sp+0]      ;get stackpointer
  ld    b, h
  ld    c, l
  ldh   a, [_TASKPTR]   ;get task pointer
  ld    l, a
  ld    h, $FF
  ld    a, b
  ld    [hl+], a        ;save stackpointer (MSB)
  ld    a, c
  ld    [hl+], a        ;save stackpointer (LSB)
  jr    .next

.skip:
  inc   l
  inc   l
  inc   l
.next:

;  bit   5, l            ;end of list? ($A0 reached)
  bit   6, l            ;end of list? ($C0 reached)
  jr    z, .norestart
  ld    l, _TASKLST     ;restart
.norestart:

  ld    a, [hl+]        ;load flags
  bit   7, a            ;valid?
  jr    z, .skip
  bit   6, a
  jr    z, .skip        ;running?

  and   a, 7
  ldh   [_SVBK], a      ;set work RAM bank

  ld    a, [hl+]
  ld    [$2000], a      ;set ROM bank

  ld    a, l
  ldh   [_TASKPTR], a   ;save new task pointer to save stackpointer

  ld    a, [hl+]        ;load stackpointer (MSB)
  ld    b, a
  ld    l, [hl]         ;load stackpointer (LSB)
  ld    h, b
  ld    sp, hl          ;set stack pointer
  popa
reti

;----------------------------------------------------------------------------;
